let bytes_of_string = String.copy
let bytes_to_string = String.copy
let bytes_make = String.make
let bytes_create = String.create
let bytes_get = String.get
let bytes_set = String.set
let bytes_length = String.length
let bytes_index_from = String.index_from

let buffer_add_subbytes = Buffer.add_substring

let string_uncapitalize = String.uncapitalize
let string_capitalize = String.capitalize
let string_lowercase = String.lowercase
let string_uppercase = String.uppercase
let char_uppercase = Char.uppercase

